from rest_framework import status
from rest_framework.test import (APIClient, APITestCase)

from util.factories import GradyUserFactory


class TutorReviewerCanChangePasswordTests(APITestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user_factory = GradyUserFactory()
        cls.data = {
            'old_password': 'l',
            'new_password': 'chompreviver0.'
        }

    def setUp(self):
        self.reviewer = self.user_factory.make_reviewer(password='l')
        self.tutor1 = self.user_factory.make_tutor(password='l')
        self.tutor2 = self.user_factory.make_tutor(password='l')
        self.client = APIClient()

    def _change_password(self, changing_user, user_to_change=None, data=None):
        if user_to_change is None:
            user_to_change = changing_user
        if data is None:
            data = self.data

        self.client.force_authenticate(user=changing_user)
        url = f"/api/user/{user_to_change.pk}/change_password/"
        return self.client.patch(url, data=data)

    def test_tutor_needs_to_provide_current_password(self):
        res = self._change_password(self.tutor1,
                                    data={'new_password': 'chompreviver0.'})
        self.assertEqual(status.HTTP_401_UNAUTHORIZED, res.status_code)
        ret = self.client.login(username=self.tutor1.username,
                                password='chompreviver0.')
        self.assertFalse(ret)

    def test_reviewer_needs_to_provide_current_password_for_self(self):
        res = self._change_password(self.reviewer,
                                    data={'new_password': 'chompreviver0.'})
        self.assertEqual(status.HTTP_401_UNAUTHORIZED, res.status_code)
        ret = self.client.login(username=self.tutor1.username,
                                password='chompreviver0.')
        self.assertFalse(ret)

    def test_tutor_can_change_own_password(self):
        res = self._change_password(self.tutor1)
        self.assertEqual(status.HTTP_200_OK, res.status_code)
        ret = self.client.login(username=self.tutor1.username,
                                password='chompreviver0.')
        self.assertTrue(ret)

    def test_tutor_cant_change_other_password(self):
        res = self._change_password(self.tutor1, self.tutor2)
        self.assertEqual(status.HTTP_403_FORBIDDEN, res.status_code)
        ret = self.client.login(username=self.tutor2.username,
                                password='chompreviver0.')
        self.assertFalse(ret)

    def test_reviewer_can_change_own_password(self):
        res = self._change_password(self.reviewer)
        self.assertEqual(status.HTTP_200_OK, res.status_code)
        ret = self.client.login(username=self.reviewer.username,
                                password='chompreviver0.')
        self.assertTrue(ret)

    def test_reviewer_can_change_tutor_password(self):
        res = self._change_password(self.reviewer, self.tutor1,
                                    data={'new_password': 'chompreviver0.'})
        self.assertEqual(status.HTTP_200_OK, res.status_code)
        ret = self.client.login(username=self.tutor1.username,
                                password='chompreviver0.')
        self.assertTrue(ret)

    def test_student_cant_change_password(self):
        student = self.user_factory.make_student(password='l')
        res = self._change_password(student)
        self.assertEqual(status.HTTP_403_FORBIDDEN, res.status_code)
        ret = self.client.login(username=student.username,
                                password='chompreviver0.')
        self.assertFalse(ret)
