""" All API views that are used to retrieve data from the database. They
can be categorized by the permissions they require. All views require a
user to be authenticated and most are only accessible by one user group """
import logging

from django.conf import settings
from django.contrib.auth.hashers import check_password
from django.db.models import Avg
import django.contrib.auth.password_validation as validators
from django.core import exceptions

from rest_framework import generics, mixins, status, viewsets
from rest_framework.decorators import (api_view, throttle_classes,
                                       action)
from rest_framework.exceptions import PermissionDenied
from rest_framework.permissions import AllowAny
from rest_framework.response import Response
from rest_framework.throttling import AnonRateThrottle

from core import models
from core.models import ExamType, StudentInfo, SubmissionType
from core.permissions import IsReviewer, IsStudent, IsTutorOrReviewer
from core.serializers import (ExamSerializer, StudentInfoSerializer,
                              StudentInfoForListViewSerializer,
                              SubmissionNoTypeSerializer, StudentSubmissionSerializer,
                              SubmissionTypeSerializer, TutorSerializer,
                              UserAccountSerializer)

log = logging.getLogger(__name__)


@api_view()
def get_jwt_expiration_delta(request):
    return Response({'timeDelta': settings.JWT_AUTH['JWT_EXPIRATION_DELTA']})


@api_view()
def get_user_role(request):
    return Response({'role': request.user.role})


class StudentSelfApiView(generics.RetrieveAPIView):
    """ Gets all data that belongs to one student """
    permission_classes = (IsStudent,)
    serializer_class = StudentInfoSerializer

    def get_object(self) -> StudentInfo:
        """ The object in question is the student associated with the requests
        user. Since the permission IsStudent is satisfied the member exists """
        if self.request.user.is_superuser:
            return StudentInfo.objects.last()
        return self.request.user.student


class StudentSelfSubmissionsApiView(generics.ListAPIView):
    permission_classes = (IsStudent, )
    serializer_class = StudentSubmissionSerializer

    def get_queryset(self):
        return self.request.user.student.submissions


class StudentReviewerApiViewSet(viewsets.ReadOnlyModelViewSet):
    """ Gets a list of all students without individual submissions """
    permission_classes = (IsReviewer,)
    queryset = StudentInfo.objects\
        .select_related('user')\
        .select_related('exam')\
        .prefetch_related('submissions')\
        .prefetch_related('submissions__feedback')\
        .prefetch_related('submissions__type')\
        .all()
    serializer_class = StudentInfoForListViewSerializer

    def _set_students_active(self, active):
        for student in self.get_queryset():
            user = student.user
            user.is_active = active
            user.save()

    @action(detail=False, methods=['post'])
    def deactivate(self, request):
        self._set_students_active(False)
        return Response(status=status.HTTP_200_OK)

    @action(detail=False, methods=['post'])
    def activate(self, request):
        self._set_students_active(True)
        return Response(status=status.HTTP_200_OK)


class ExamApiViewSet(viewsets.ReadOnlyModelViewSet):
    """ Gets a list of an individual exam by Id if provided """
    permission_classes = (IsReviewer,)
    queryset = ExamType.objects.all()
    serializer_class = ExamSerializer


class TutorApiViewSet(
        mixins.RetrieveModelMixin,
        mixins.UpdateModelMixin,
        mixins.CreateModelMixin,
        mixins.DestroyModelMixin,
        mixins.ListModelMixin,
        viewsets.GenericViewSet):
    """ Api endpoint for creating, listing, viewing or deleting tutors """
    permission_classes = (IsReviewer,)
    queryset = models.UserAccount.tutors \
        .with_feedback_count() \
        .prefetch_related('subscriptions') \
        .prefetch_related('subscriptions__assignments')
    serializer_class = TutorSerializer

    @action(detail=False, methods=['post'], permission_classes=[AllowAny])
    @throttle_classes([AnonRateThrottle])
    def register(self, request):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        if serializer.validated_data.get('is_active', False):
            raise PermissionDenied(detail='Cannot be created active')

        serializer.save()
        return Response(serializer.data, status=status.HTTP_201_CREATED)


class SubmissionTypeApiView(viewsets.ReadOnlyModelViewSet):
    """ Gets a list or a detail view of a single SubmissionType """
    queryset = SubmissionType.objects.all()
    serializer_class = SubmissionTypeSerializer


class StatisticsEndpoint(viewsets.ViewSet):

    def list(self, request, *args, **kwargs):
        first_sub_type = models.SubmissionType.objects.first()

        return Response({
            'submissions_per_type':
                first_sub_type.submissions.count() if first_sub_type is not None else 0,

            'submissions_per_student':
            models.SubmissionType.objects.count(),

            'current_mean_score':
            models.Feedback.objects.aggregate(avg=Avg('score')).get('avg', 0),

            'submission_type_progress':
                # Queryset is explicitly evaluated so camelizer plugin camelizes it
                list(models.SubmissionType.get_annotated_feedback_count().values(
                    'pk',
                    'name',
                    'submission_count',
                    'feedback_final',
                    'feedback_in_validation',
                    'feedback_in_conflict'))
        })


class SubmissionViewSet(viewsets.ReadOnlyModelViewSet):
    permission_classes = (IsTutorOrReviewer, )
    serializer_class = SubmissionNoTypeSerializer
    queryset = models.Submission.objects\
        .select_related('type')\
        .select_related('feedback')\
        .prefetch_related('tests')\
        .prefetch_related('feedback__feedback_lines')\
        .prefetch_related('feedback__feedback_lines__of_tutor')\
        .all()

    def get_queryset(self):
        if self.request.user.is_reviewer():
            return self.queryset

        return self.queryset.filter(
            assignments__subscription__owner=self.request.user
        )


class UserAccountViewSet(viewsets.ReadOnlyModelViewSet):
    serializer_class = UserAccountSerializer
    queryset = models.UserAccount.objects.all()

    @action(detail=True, methods=['patch'], permission_classes=(IsTutorOrReviewer, ))
    def change_password(self, request, *args, **kwargs):
        user = self.get_object()
        if request.user != user and not request.user.is_reviewer():
            return Response(status=status.HTTP_403_FORBIDDEN)
        old_password = request.data.get('old_password')

        # tutors must always provide their current password
        # reviewers must provide their current password when they change
        # their own, not if they change the password of a tutor
        if (request.user.is_tutor() or
            request.user.is_reviewer and request.user == user) \
            and \
            (old_password is None or
                not check_password(old_password, user.password)):
                    return Response(status=status.HTTP_401_UNAUTHORIZED)

        new_password = request.data.get('new_password')
        # validate password
        try:
            if new_password is not None:
                validators.validate_password(password=new_password, user=user)
        except exceptions.ValidationError as err:
            return Response({'new_password': list(err.messages)},
                            status=status.HTTP_406_NOT_ACCEPTABLE)
        user.set_password(new_password)
        user.save()
        log.info(f"User {request.user} changed password of {user}")
        return Response(status=status.HTTP_200_OK)

    @action(detail=True, methods=['patch'])
    def change_active(self, request, *args, **kwargs):
        active = request.data.get('is_active')
        req_user = request.user
        user = self.get_object()
        if active is None:
            error_msg = "You need to provide an 'active' field"
            return Response({'Error': error_msg}, status.HTTP_400_BAD_REQUEST)
        if (req_user.is_student() or req_user.is_tutor()) and req_user != user:
            return Response(status.HTTP_403_FORBIDDEN)
        user.is_active = active
        user.save()
        return Response(status.HTTP_200_OK)

    @action(detail=False)
    def me(self, request):
        serializer = self.get_serializer(request.user)
        return Response(serializer.data, status=status.HTTP_200_OK)
